
from dataclasses import asdict
from datetime import datetime, time, timedelta
from zoneinfo import ZoneInfo

import pandas as pd
import streamlit as st
from matplotlib import pyplot as plt

from minitrade.backtest import calculate_trade_stats
from minitrade.broker import Broker, BrokerAccount
from minitrade.datasource import QuoteSource
from minitrade.trader import (BacktestLog, BacktestRunner, StrategyManager,
                              TradePlan)
from minitrade.utils.mtdb import MTDB

st.set_page_config(page_title='Trading', layout='wide')


def market_timezone_selectbox():
    major_market_timezones = ['America/New_York', 'Asia/Shanghai', 'Asia/Hong_Kong']
    return st.selectbox('Select market local timezone', options=major_market_timezones)


def ticker_resolver(account: BrokerAccount, ticker_css: str) -> dict:
    ''' Map ticker to broker specific contract ID '''
    if account and len(ticker_css.replace(' ', '')) > 0:
        broker = Broker.get_broker(account)
        with st.expander('Resolve tickers', expanded=True):
            try:
                broker.connect()
                candidates: dict = broker.resolve_tickers(ticker_css.replace(' ', ''))
                if candidates:
                    vendor_tickers = {}
                    for k, v in candidates.items():
                        vendor_tickers[k] = st.radio(k, v, format_func=lambda _: _['label'])
                    return vendor_tickers
            except ConnectionError:
                st.error('Need to login to broker account to resolve tickers. Please pay attention to 2FA notification if enabled.')
                st.button('Retry')
    return None


def get_broker_ticker_map(tickers: dict | None) -> dict | None:
    ''' Verify if all tickers have been resolved and return the mapping '''
    if tickers and None not in tickers.values():
        return {k: v['id'] for k, v in tickers.items()}
    else:
        return None


def show_create_trade_plan_form() -> TradePlan | None:
    strategy_file = st.selectbox('Pick a strategy', StrategyManager.list())
    ticker_css = st.text_input(
        'Define the asset space (a list of tickers separated by comma without space)', placeholder='AAPL,GOOG')
    data_source = st.selectbox('Select a data source', QuoteSource.AVAILABLE_SOURCES)
    market_timezone = market_timezone_selectbox()
    backtest_start_date = st.date_input(
        'Pick a backtest start date (run backtest from that date to give enough lead time to calculate indicators)',
        value=datetime.today() - timedelta(days=110))
    trade_start_date = st.date_input(
        'Pick a trade start date (trade signal before that is surpressed)', min_value=datetime.now().date())
    trade_time_of_day = st.time_input('Pick when backtesting should start (market local time)', value=time(19, 30))
    account = st.selectbox('Select a broker account', BrokerAccount.list(), format_func=lambda b: b.alias)
    initial_cash = st.number_input('Set the cash amount to invest', value=0)
    initial_holding = st.text_input(
        'Set the preexisting asset positions (number of shares you already own and to be considered in the strategy)',
        placeholder='AAPL:100,GOOG:100') or None
    name = st.text_input('Name the trade plan')
    tickers = ticker_resolver(account, ticker_css)
    dryrun = st.button('Save and dry run')
    if dryrun:
        ticker_map = get_broker_ticker_map(tickers)
        if initial_holding:
            initial_holding = {k.strip(): int(v)
                               for k, v in [x.strip().split(':') for x in initial_holding.split(',')]}
        if len(ticker_css.strip()) == 0 or len(name.strip()) == 0:
            st.error('Please do not leave any required field empty.')
        elif account and ticker_map is None:
            st.error('Tickers are not fully resolved.')
        else:
            return TradePlan(
                id=MTDB.uniqueid(),
                name=name,
                strategy_file=strategy_file,
                ticker_css=ticker_css.replace(' ', ''),
                market_timezone=market_timezone,
                data_source=data_source,
                backtest_start_date=backtest_start_date.strftime('%Y-%m-%d'),
                trade_start_date=trade_start_date.strftime('%Y-%m-%d'),
                trade_time_of_day=trade_time_of_day.strftime('%H:%M:%S'),
                broker_account=account.alias if account else None,
                initial_cash=initial_cash,
                initial_holding=initial_holding,
                enabled=False,
                create_time=datetime.utcnow(),
                update_time=None,
                broker_ticker_map=ticker_map
            )


def display_run(plan: TradePlan, log: BacktestLog):
    orders = plan.get_orders(log.id)
    log_status = '❌' if log.error else '✅' if orders else '🟢'
    label = f'{log_status} {log.log_time} [{log.id}]' + (f' **{len(orders)} orders**' if orders else '')
    with st.expander(label):
        tab1, tab2, tab3, tab4 = st.tabs(['Result', 'Error', 'Log', 'Orders'])
        if log.result is not None:
            df = log.result
            df = df[~df.index.str.startswith('_')].T
            tab1.write(df)
        tab2.code(log.exception)
        tab3.caption('Log - stdout')
        if log.stdout:
            tab3.text(log.stdout)
        tab3.caption('Log - stderr')
        if log.stderr:
            tab3.text(log.stderr)
        if orders:
            tab4.write(pd.DataFrame(orders))


def save_plan_and_dryrun(plan: TradePlan) -> None:
    plan.save()
    runner = BacktestRunner(plan)
    log = runner.execute(dryrun=True)
    display_run(plan, log)
    if log.error:
        st.error('Trade plan dryrun failed, plan disabled')
    else:
        plan.enable(True)
        st.success('Trade plan dryrun succeeded, plan enabled')


def show_trade_plan_selector() -> TradePlan | None:
    plan_lst = TradePlan.list_plans()
    plan_idx = st.sidebar.radio('Trade plan', list(range(len(plan_lst))),
                                format_func=lambda i: plan_lst[i].name)
    plan = plan_lst[int(plan_idx)] if plan_idx is not None else None
    return plan


def run_trade_plan_once(plan: TradePlan) -> None:
    runner = BacktestRunner(plan)
    log = runner.execute()
    if log is not None and not log.error:
        st.success(f'Backtest {log.id} finished successfully')
    else:
        st.error(f'Backtest failed')


def confirm_delete_trade_plan(plan: TradePlan) -> None:
    def confirm_delete():
        if st.session_state.delete_confirm_textinput == plan.name:
            plan.delete()
    st.text_input(f'Type "{plan.name}" and press Enter to delete',
                  on_change=confirm_delete, key='delete_confirm_textinput')


def show_trade_plan_header_and_controls(plan: TradePlan) -> None:
    c1, c2, c3, c4, c5 = st.columns([5, 1, 1, 1, 1])
    c1.subheader(plan.name)
    c2.button('Refresh')
    c3.button('Run once', on_click=lambda: run_trade_plan_once(plan))
    c4.button('Disable' if plan.enabled else 'Enable', key='tradeplan_onoff',
              on_click=lambda: plan.enable(not plan.enabled))
    if c5.button('Delete', type='primary'):
        confirm_delete_trade_plan(plan)


def show_trade_plan_status(plan: TradePlan) -> None:
    job = plan.jobinfo()
    if job is None:
        if plan.enabled:
            st.warning('State inconsistency detected: plan is enabled but not scheduled')
        else:
            st.warning('Plan is disabled and not trading')
    else:
        if plan.enabled:
            st.success(f'Plan is scheduled to run at {job["next_run_time"]}')
        else:
            st.success(
                f'State inconsistenccy detected: plan is disabled but is scheduled to run at {job["next_run_time"]}')


def show_trade_plan_execution_history(plan: TradePlan) -> None:
    tab1, tab2, tab3, tab4 = st.tabs(['Run history', 'Orders', 'Performance', 'Settings'])
    logs = plan.list_logs()
    with tab1:
        show_trade_plan_status(plan)
        for log in logs:
            display_run(plan, log)

    with tab2:
        account = BrokerAccount.get_account(plan)
        broker = account and Broker.get_broker(account)
        orders = plan.get_orders()
        for order in orders:
            trade = broker and broker.find_trades(order)
            broker_order = broker and broker.find_order(order)
            order_status = '✅' if trade else '🟢' if order.broker_order_id else '🚫'
            with st.expander(f'{order_status} {order.signal_time} **{order.ticker} {order.side} {abs(order.size)}**'):
                t1, t2, t3 = st.tabs(['Raw order', 'Broker order', 'Trade status'])
                t1.dataframe(pd.DataFrame([asdict(order)]))
                if broker_order:
                    t2.dataframe(pd.DataFrame([broker_order]))
                if trade:
                    t3.dataframe(pd.DataFrame(trade))
                    if account.broker == 'Manual':
                        c1, c2, c3, c4 = t3.columns(4)
                        price = c1.number_input('Price', key=f'{order.id}-price', value=trade[0]['price'] or 0.0)
                        commission = c2.number_input(
                            'Commission', key=f'{order.id}-comm', value=trade[0]['commission'] or 0.0)
                        dt = c3.date_input('Trade time (market local timezone)',
                                           key=f'{order.id}-date', value=trade[0]['trade_time'])
                        tm = c4.time_input('Trade time', key=f'{order.id}-time',
                                           value=trade[0]['trade_time'] or time(9, 30), label_visibility='hidden')
                        if t3.button('Save', key=f'{order.id}-save'):
                            broker.update_trade(trade[0]['id'], price, commission,
                                                datetime.combine(dt, tm, ZoneInfo(plan.market_timezone)))
                            st.warning('Click "Refresh" button to see change')

    with tab3:
        if broker and logs:
            try:
                data = logs[0].data
                trades = broker.format_trades(orders)
                trade_start_date = datetime.strptime(
                    plan.trade_start_date, '%Y-%m-%d').replace(tzinfo=ZoneInfo(plan.market_timezone))
                offset = max((data.index >= trade_start_date).sum(), 1)
                # calculate trade stats from trade_start_date
                trade_df, equity, pnl, commission_rate = calculate_trade_stats(
                    data.iloc[-offset:], plan.initial_cash, trades, plan.initial_holding)
                rr = (equity / equity[0] - 1) * 100
                rr.name = 'Return rate (%)'
                c1, c2, c3, c4 = st.columns(4)
                c1.metric(
                    label='Portfolio Value', value=f'{equity.iloc[-1]:,.0f}',
                    delta=f'{int(equity.iloc[-1]-equity.iloc[-2]):,}' if len(equity) >= 2 else 0)
                c2.metric(label='PnL', value=f'{(equity.iloc[-1]-equity.iloc[0]):,.0f}')
                c3.metric(label='Return', value=f'{rr.iloc[-1]:.2f}%')
                c4.metric(label='Commission Rate', value=f'{commission_rate:.3%}')
                if len(rr) >= 2:
                    st.caption('Return rate (%)')
                    fig, ax = plt.subplots()
                    rr.plot(kind='line', ax=ax, xticks=rr.index, xlabel='date')
                    st.pyplot(fig)
                st.caption('Profit and loss')
                st.write(pnl.style.format('{:,.0f}', na_rep=' '))
                st.caption('Trades')
                st.write(trade_df.style.format(na_rep=' '))
            except Exception as e:
                st.write(e)

    with tab4:
        st.write(asdict(plan))


action = st.sidebar.radio('Action', ['Browse existing', 'Create new'])

if action == 'Create new':
    st.subheader('Create new trade plan')
    plan = show_create_trade_plan_form()
    if plan:
        save_plan_and_dryrun(plan)

if action == 'Browse existing':
    plan = show_trade_plan_selector()
    if plan:
        show_trade_plan_header_and_controls(plan)
        show_trade_plan_execution_history(plan)



import warnings
from typing import TYPE_CHECKING, List, Union

import numpy as np
import pandas as pd

from ._util import _data_period

if TYPE_CHECKING:
    from .backtesting import Order, Strategy, Trade


def compute_drawdown_duration_peaks(dd: pd.Series):
    iloc = np.unique(np.r_[(dd == 0).values.nonzero()[0], len(dd) - 1])
    iloc = pd.Series(iloc, index=dd.index[iloc])
    df = iloc.to_frame('iloc').assign(prev=iloc.shift())
    df = df[df['iloc'] > df['prev'] + 1].astype(int)

    # If no drawdown since no trade, avoid below for pandas sake and return nan series
    if not len(df):
        return (dd.replace(0, np.nan),) * 2

    df['duration'] = df['iloc'].map(dd.index.__getitem__) - df['prev'].map(dd.index.__getitem__)
    df['peak_dd'] = df.apply(lambda row: dd.iloc[row['prev']:row['iloc'] + 1].max(), axis=1)
    df = df.reindex(dd.index)
    return df['duration'], df['peak_dd']


def geometric_mean(returns: pd.Series) -> float:
    returns = returns.fillna(0) + 1
    if np.any(returns <= 0):
        return 0
    return np.exp(np.log(returns).sum() / (len(returns) or np.nan)) - 1


def compute_stats(
        orders: Union[List['Order'], pd.DataFrame],
        trades: Union[List['Trade'], pd.DataFrame],
        equity: pd.DataFrame,
        ohlc_data: pd.DataFrame,
        strategy_instance: 'Strategy',
        risk_free_rate: float = 0,
) -> pd.Series:
    assert -1 < risk_free_rate < 1

    index = ohlc_data.index
    dd = 1 - equity['_Equity'] / np.maximum.accumulate(equity['_Equity'])
    dd_dur, dd_peaks = compute_drawdown_duration_peaks(pd.Series(dd, index=index))

    if isinstance(orders, pd.DataFrame):
        orders_df = orders
    else:
        orders_df = pd.DataFrame({
            'SignalTime': [t.entry_time for t in orders],
            'Ticker': [t.ticker for t in orders],
            'Side': ['Buy' if t.size > 0 else 'Sell' for t in orders],
            'Size': [int(t.size) for t in orders],
            'EntryType': [t.entry_type for t in orders],
        }).set_index('SignalTime')

    equity_df = pd.concat([equity, pd.DataFrame({'DrawdownPct': dd, 'DrawdownDuration': dd_dur}, index=index)], axis=1)

    if isinstance(trades, pd.DataFrame):
        trades_df = trades
    else:
        # Came straight from Backtest.run()
        trades_df = pd.DataFrame({
            'EntryBar': [t.entry_bar for t in trades],
            'ExitBar': [t.exit_bar for t in trades],
            'Ticker': [t.ticker for t in trades],
            'Size': [t.size for t in trades],
            'EntryPrice': [t.entry_price for t in trades],
            'ExitPrice': [t.exit_price for t in trades],
            'PnL': [t.pl for t in trades],
            'ReturnPct': [t.pl_pct for t in trades],
            'EntryTime': [t.entry_time for t in trades],
            'ExitTime': [t.exit_time for t in trades],
            'Tag': [t.tag for t in trades],
        })
        trades_df['Duration'] = trades_df['ExitTime'] - trades_df['EntryTime']
    del trades

    pl = trades_df['PnL']
    returns = trades_df['ReturnPct']
    durations = trades_df['Duration']

    def _round_timedelta(value, _period=_data_period(index)):
        if not isinstance(value, pd.Timedelta):
            return value
        resolution = getattr(_period, 'resolution_string', None) or _period.resolution
        return value.ceil(resolution)

    s = pd.Series(dtype=object)
    s.loc['Start'] = index[0]
    s.loc['End'] = index[-1]
    s.loc['Duration'] = s.End - s.Start

    have_position = np.repeat(0, len(index))
    for t in trades_df.itertuples(index=False):
        have_position[t.EntryBar:t.ExitBar + 1] = 1

    s.loc['Exposure Time [%]'] = have_position.mean() * 100  # In "n bars" time, not index time
    s.loc['Equity Final [$]'] = equity['_Equity'].iloc[-1]
    s.loc['Equity Peak [$]'] = equity['_Equity'].max()
    s.loc['Return [%]'] = (equity['_Equity'].iloc[-1] - equity['_Equity'].iloc[0]) / equity['_Equity'].iloc[0] * 100
    c = ohlc_data.Close.values
    s.loc['Buy & Hold Return [%]'] = (c[-1] - c[0]) / c[0] * 100  # long-only return

    gmean_day_return: float = 0
    day_returns = np.array(np.nan)
    annual_trading_days = np.nan
    if isinstance(index, pd.DatetimeIndex):
        day_returns = equity_df['_Equity'].resample('D').last().dropna().pct_change()
        gmean_day_return = geometric_mean(day_returns)
        annual_trading_days = float(
            365 if index.dayofweek.to_series().between(5, 6).mean() > 2/7 * .6 else 252)

    # Annualized return and risk metrics are computed based on the (mostly correct)
    # assumption that the returns are compounded. See: https://dx.doi.org/10.2139/ssrn.3054517
    # Our annualized return matches `empyrical.annual_return(day_returns)` whereas
    # our risk doesn't; they use the simpler approach below.
    annualized_return = (1 + gmean_day_return)**annual_trading_days - 1
    s.loc['Return (Ann.) [%]'] = annualized_return * 100
    s.loc['Volatility (Ann.) [%]'] = np.sqrt((day_returns.var(ddof=int(bool(day_returns.shape))) + (1 + gmean_day_return)**2)**annual_trading_days - (1 + gmean_day_return)**(2*annual_trading_days)) * 100  # noqa: E501
    # s.loc['Return (Ann.) [%]'] = gmean_day_return * annual_trading_days * 100
    # s.loc['Risk (Ann.) [%]'] = day_returns.std(ddof=1) * np.sqrt(annual_trading_days) * 100

    # Our Sharpe mismatches `empyrical.sharpe_ratio()` because they use arithmetic mean return
    # and simple standard deviation
    s.loc['Sharpe Ratio'] = (s.loc['Return (Ann.) [%]'] - risk_free_rate) / (s.loc['Volatility (Ann.) [%]'] or np.nan)  # noqa: E501
    with warnings.catch_warnings():
        # wrap to catch RuntimeWarning: divide by zero encountered in scalar divide
        warnings.filterwarnings('error')
        try:
            # Our Sortino mismatches `empyrical.sortino_ratio()` because they use arithmetic mean return
            s.loc['Sortino Ratio'] = (annualized_return - risk_free_rate) / (np.sqrt(np.mean(day_returns.clip(-np.inf, 0)**2)) * np.sqrt(annual_trading_days))  # noqa: E501
        except Warning:
            s.loc['Sortino Ratio'] = np.nan
    max_dd = -np.nan_to_num(dd.max())
    s.loc['Calmar Ratio'] = annualized_return / (-max_dd or np.nan)
    s.loc['Max. Drawdown [%]'] = max_dd * 100
    s.loc['Avg. Drawdown [%]'] = -dd_peaks.mean() * 100
    s.loc['Max. Drawdown Duration'] = _round_timedelta(dd_dur.max())
    s.loc['Avg. Drawdown Duration'] = _round_timedelta(dd_dur.mean())
    s.loc['# Trades'] = n_trades = len(trades_df)
    win_rate = np.nan if not n_trades else (pl > 0).mean()
    s.loc['Win Rate [%]'] = win_rate * 100
    s.loc['Best Trade [%]'] = returns.max() * 100
    s.loc['Worst Trade [%]'] = returns.min() * 100
    mean_return = geometric_mean(returns)
    s.loc['Avg. Trade [%]'] = mean_return * 100
    s.loc['Max. Trade Duration'] = _round_timedelta(durations.max())
    s.loc['Avg. Trade Duration'] = _round_timedelta(durations.mean())
    s.loc['Profit Factor'] = returns[returns > 0].sum() / (abs(returns[returns < 0].sum()) or np.nan)  # noqa: E501
    s.loc['Expectancy [%]'] = returns.mean() * 100
    s.loc['SQN'] = np.sqrt(n_trades) * pl.mean() / (pl.std() or np.nan)
    s.loc['Kelly Criterion'] = win_rate - (1 - win_rate) / (pl[pl > 0].mean() / -pl[pl < 0].mean())

    s.loc['_strategy'] = strategy_instance
    s.loc['_equity_curve'] = equity_df
    s.loc['_trades'] = trades_df
    s.loc['_orders'] = orders_df

    s = _Stats(s)
    return s


class _Stats(pd.Series):
    def __repr__(self):
        # Prevent expansion due to _equity and _trades dfs
        with pd.option_context('max_colwidth', 20):
            return super().__repr__()



import hashlib
import os
import pkgutil
import platform
import sqlite3
import subprocess
import sys
from posixpath import expanduser
from zipfile import ZipFile

import click
import requests
import yaml
from tqdm import tqdm


def __call_ibgateway_admin(method: str, path: str, params: dict | None = None):
    '''Call the ibgateway's admin API

    Parameters
    ----------
    method : str
        The HTTP method, i.e. GET, POST, PUT, DELETE, etc.
    path : str
        The REST API endpoint, e.g. '/ibgateway'
    params : dict, optional
        Extra parameters to be sent along with the REST API call

    Returns
    -------
    Any
        result json if API returns 200, otherwise None
    '''
    from minitrade.utils.config import config
    url = f'http://{config.brokers.ib.gateway_admin_host}:{config.brokers.ib.gateway_admin_port}{path}'
    resp = requests.request(method=method, url=url, params=params)
    if resp.status_code == 200:
        return resp.json()
    elif resp.status_code >= 400:
        raise RuntimeError(f'Request {path} returned {resp.status_code} {resp.text}')


@click.group()
@click.pass_context
def mtcli(ctx):
    pass


@mtcli.group()
def ib():
    '''Control the execution of IB gateway'''
    pass


@ib.command('start')
def ib_start():
    import uvicorn

    from minitrade.utils.config import config
    uvicorn.run(
        'minitrade.broker.ibgateway:app',
        host=config.brokers.ib.gateway_admin_host,
        port=config.brokers.ib.gateway_admin_port,
        log_level=config.brokers.ib.gateway_admin_log_level,
    )


@ib.command('status')
def ib_status():
    status = __call_ibgateway_admin('GET', '/ibgateway')
    click.echo(status)


@ib.command('login')
@click.option('-a', '--alias', prompt='Alias', help='IB account alias')
def ib_login(alias):
    status = __call_ibgateway_admin('PUT', f'/ibgateway/{alias}')
    click.echo(status)


@ib.command('logout')
def ib_logout():
    accounts = __call_ibgateway_admin('GET', '/ibgateway')
    for a in accounts:
        click.echo(a)
        __call_ibgateway_admin('DELETE', f'/ibgateway/{a["account"]}')


@mtcli.command()
def web():
    '''Launch web UI'''
    from streamlit.web.cli import main
    cd = os.path.dirname(os.path.realpath(__file__))
    home = os.path.join(cd, 'admin/Home.py')
    lib_home = os.path.join(cd, '/..')
    os.chdir(lib_home)
    sys.argv = ["streamlit", "run", home]
    sys.exit(main())


@mtcli.group()
def scheduler():
    '''Manage scheduler and scheduled jobs'''
    pass


def __call_scheduler(method: str, path: str, params: dict | None = None):
    '''Call the scheduler's REST API

    Parameters
    ----------
    method : str
        The HTTP method, i.e. GET, POST, PUT, DELETE, etc.
    path : str
        The REST API endpoint, e.g. '/jobs'
    params : dict, optional
        Extra parameters to be sent along with the REST API call

    Returns
    -------
    Any
        result json if API returns 200, otherwise None
    '''
    from minitrade.utils.config import config
    url = f'http://{config.scheduler.host}:{config.scheduler.port}{path}'
    resp = requests.request(method=method, url=url, params=params)
    if resp.status_code == 200:
        return resp.json()


@scheduler.command('start')
def scheduler_start():
    import uvicorn

    from minitrade.utils.config import config
    uvicorn.run(
        'minitrade.trader.scheduler:app',
        host=config.scheduler.host,
        port=config.scheduler.port,
        log_level=config.scheduler.log_level,
    )


@scheduler.command('status')
def scheduler_status():
    status = __call_scheduler('GET', '/jobs')
    click.echo(status)


@scheduler.command('schedule')
@click.argument('plan_id')
def scheduler_schedule(plan_id):
    status = __call_scheduler('PUT', f'/jobs/{plan_id}')
    click.echo(status)


@scheduler.command('unschedule')
@click.argument('plan_id')
def scheduler_unschedule(plan_id):
    status = __call_scheduler('DELETE', f'/jobs/{plan_id}')
    click.echo(status)


@mtcli.command
@click.argument('plan_id')
@click.option('--run_id', default=None, help='Specify an unique run ID manually')
@click.option('--dryrun', is_flag=True, help='Dry run only, don\'t place orders')
@click.option('--pytest', is_flag=True, help='Run in test mode')
def backtest(plan_id, run_id, dryrun, pytest):
    '''Run backtest for particular plan'''
    from minitrade.trader.trader import BacktestRunner
    if pytest:
        # import pytest which will trigger running in test mode using test db
        import pytest

    from minitrade.trader import TradePlan
    try:
        plan = TradePlan.get_plan(plan_id)
        runner = BacktestRunner(plan)
        runner.run_backtest(run_id=run_id, dryrun=dryrun)
    except Exception as e:
        raise e


def check_program_version(name):
    proc = subprocess.run(['which', name], capture_output=True, cwd=os.getcwd(), text=True)
    if proc.returncode == 0:
        path = proc.stdout.strip('\n')
        proc = subprocess.run([path, '--version'], capture_output=True, cwd=os.getcwd(), text=True)
        if proc.returncode == 0:
            click.secho(f'  {name:15s} ... found', fg='green')
            return True
    click.secho(f'  {name:15s} ... missing', fg='red')
    return False


def check_selenium():
    try:
        from selenium import webdriver
        options = webdriver.ChromeOptions()
        options.add_argument('ignore-certificate-errors')
        options.add_argument("--headless")
        with webdriver.Chrome(options=options) as driver:
            if driver.capabilities:
                click.secho(f'  {"selenium":15s} ... working', fg='green')
                return True
    except Exception as e:
        click.secho(f'  selenium ... error: {e}', fg='red')
        return False


@mtcli.command()
def init():
    '''Initialize Minitrade for trading'''
    if platform.system() == 'Linux':
        click.secho('Checking prerequisites:')
        prerequisites = [
            check_program_version('java'),
            check_program_version('google-chrome'),
            check_selenium()
        ]
        if not all(prerequisites):
            click.secho('Please install the required dependencies.', fg='red')
            sys.exit(1)

    if os.path.exists(expanduser('~/.minitrade')):
        click.secho('Error: ~/.minitrade directory already exists. Initialization aborted.', fg='red')
        click.secho(
            'If you want to reset the Minitrade installation, please delete ~/.minitrade directory and try again.',
            fg='red')
        sys.exit(1)

    minitrade_root = expanduser('~/.minitrade')
    # init dirs
    click.secho(f'Setting up directories...')
    os.makedirs(os.path.join(minitrade_root, 'database'), mode=0o700, exist_ok=True)
    os.makedirs(os.path.join(minitrade_root, 'strategy'), mode=0o700, exist_ok=True)
    os.makedirs(os.path.join(minitrade_root, 'ibgateway'), mode=0o700, exist_ok=True)
    # init config
    from minitrade.utils.config import GlobalConfig
    GlobalConfig().save()
    # init db
    db_loc = os.path.join(minitrade_root, 'database/minitrade.db')
    click.secho(f'Setting up database...')
    sql = pkgutil.get_data(__name__, 'minitrade.db.sql').decode('utf-8')
    with sqlite3.connect(db_loc) as conn:
        conn.executescript(sql)
    conn.close()
    # populate nasdaq tickers
    from minitrade.datasource import populate_nasdaq_traded_symbols
    populate_nasdaq_traded_symbols()
    # download and extract IB gateway
    click.secho(f'Installing Interactive Brokers gateway...')
    ib_loc = os.path.join(minitrade_root, 'ibgateway')
    url = 'https://download2.interactivebrokers.com/portal/clientportal.gw.zip'
    response = requests.get(url, stream=True)
    total_kb = int(int(response.headers["Content-Length"]) / 1000)
    with open(f'{ib_loc}/clientportal.gw.zip', "wb") as f:
        file_hash = hashlib.md5()
        t = tqdm(response.iter_content(chunk_size=1000), total=total_kb,
                 unit="KB", desc='Downloading IB gateway', leave=False)
        for data in t:
            file_hash.update(data)
            f.write(data)
            t.format_sizeof(len(data))
        t.close()
        with open(f'{ib_loc}/clientportal.gw.zip.md5', "w") as h:
            h.write(file_hash.hexdigest())
    ZipFile(f'{ib_loc}/clientportal.gw.zip').extractall(ib_loc)
    # tighten API access to localhost only
    try:
        conf_loc = os.path.join(ib_loc, 'root/conf.yaml')
        with open(conf_loc, 'r') as f:
            conf = yaml.safe_load(f)
        conf['ips']['allow'] = ['127.0.0.1']
        conf['ips']['deny'] = []
        with open(conf_loc, 'w') as f:
            yaml.safe_dump(conf, f)
    except Exception as e:
        raise RuntimeError(f'Writing gateway config file failed: {conf_loc}') from e
    # finish
    click.secho(f'Minitrade initialized in {minitrade_root}', fg='green')


if __name__ == '__main__':
    mtcli()



import streamlit as st

from minitrade.broker import Broker, BrokerAccount

st.set_page_config(page_title='Broker', layout='wide')


def show_create_broker_account_form():
    brokers = Broker.AVAILABLE_BROKERS
    broker = st.selectbox('Broker', brokers.keys(), format_func=lambda k: brokers[k])
    mode = st.radio('Account type', ['Paper', 'Live'])
    username = st.text_input('Username')
    password = st.text_input('Password', type='password')
    alias = st.text_input('Alias', value=f'{broker}-{mode}-{username}')
    save = st.button('Test and save')
    if save:
        return BrokerAccount(
            alias=alias,
            broker=broker,
            mode=mode,
            username=username,
            password=password,
        )


def test_and_save_broker_account(account: BrokerAccount):
    if account:
        account.save()
        broker = Broker.get_broker(account)
        st.info(
            f'Logging in account {account.username}. Please pay attention to 2FA notification if enabled.')
        try:
            broker.connect()
            st.success('Login succeeded')
        except ConnectionError:
            st.error('Login failed')


def show_broker_account_selector():
    acc_lst = BrokerAccount.list()
    acc_idx = st.sidebar.radio('Broker account', list(range(len(acc_lst))),
                               format_func=lambda i: acc_lst[i].alias)
    account = acc_lst[int(acc_idx)] if acc_idx is not None else None
    return account


def confirm_delete_broker_account(account: BrokerAccount) -> None:
    def confirm_delete():
        if st.session_state.delete_confirm_textinput == account.alias:
            account.delete()
    st.text_input(f'Type "{account.alias}" and press Enter to delete',
                  on_change=confirm_delete, key='delete_confirm_textinput')


def show_broker_account_header_and_controls(account: BrokerAccount):
    c1, c2, c3, c4 = st.columns([6, 1, 1, 1])
    c1.subheader(account.alias)
    c2.button('Refresh')
    if c3.button('Login'):
        broker = Broker.get_broker(account)
        try:
            broker.connect()
            st.success('Login succeeded')
        except ConnectionError:
            st.error('Login failed')
    if c4.button('Delete', type='primary'):
        confirm_delete_broker_account(account)


def show_broker_account_portfolio_and_trades(account: BrokerAccount):
    tab1, tab2, tab3, tab4 = st.tabs(['Portfolio', 'Recent trades', 'Orders', 'Account info'])
    broker = Broker.get_broker(account)
    if broker.is_ready():
        info = broker.get_account_info()
        tab1.write(broker.get_portfolio())
        tab2.write(broker.download_trades())
        tab3.write(broker.download_orders())
        tab4.write(info)
    else:
        st.info('Login to see account info')


action = st.sidebar.radio('Action', ['Browse existing', 'Add new'])

if action == 'Add new':
    st.subheader('Add broker account')
    account = show_create_broker_account_form()
    if account:
        test_and_save_broker_account(account)

if action == 'Browse existing':
    account = show_broker_account_selector()
    if account:
        show_broker_account_header_and_controls(account)
        show_broker_account_portfolio_and_trades(account)



import sys
import unittest

suite = unittest.defaultTestLoader.discover('minitrade.backtest.core.test',
                                            pattern='_test*.py', top_level_dir='minitrade')
if __name__ == '__main__':
    result = unittest.TextTestRunner(verbosity=2).run(suite)
    sys.exit(not result.wasSuccessful())
